'Graphics Demo for QMP, shows how to display bitmap images and also graphics primitives
'Copyright (C) 2013 Raymond Allen.
'  MIT License:   See end of file for terms of use. 

'Notes:

CON
  _CLKMODE      = XTAL1 + PLL16X                       
  _XINFREQ      = 5_000_000

CON 'SD card pin settings      Make sure matches your setup!
  'Propeller Platform Express settings:
  SD_MISO=27
  SD_Clk=26
  SD_MOSI=25
  SD_CS=24

OBJ
  lcd : "QMP_DisplayDriver2d"  'Controls the LCD
  sd : "fsrw"
   

VAR  'Display variables
  long foregroundcolor,backgroundcolor

  
VAR 'buffer for bitmap line of pixels
  long lineBuffer[320*3/4]' buffer for line of an image and scrolling  'needs to be long aligned
  long BmpPalette[256]
      
pub Main               'Startup Routine

  outa[9..10]~~
  dira[9..10]~~
  

  'initialize the LCD
  lcd.start
  
  'Show Demo
  repeat
    Text_Demo
    Font_Demo
    Scroll_Demo  'Note: scroll is slow in landscape mode due to no hardware support 
    Gfx_Demo


PUB Text_Demo|i,j,k 'copies functionality of VGA_Text_Demo.Spin
  'Slightly modified to account for different font being used
  'Note:  This is not VGA Text Demo, is not 2.6kB, and is not using Internal ROM font!
  lcd.StartText(@font2,@linebuffer)   'with a 8x12 font from VGA_HIRES_TEXT (but only characters 0..127)
  lcd.out(0) 'clear screen
  
  lcd.str(string(13,"   VGA Text Demo...",13,13,$C,5," OBJ and VAR require only 2.6KB ",$C,1))
  repeat 14
    lcd.out(" ")
  repeat 4
    repeat i from $0E to 127
      lcd.out(i)
  lcd.str(string($C,6,"     Uses internal ROM font     ",$C,2))
  repeat 3000 'slight mod
    lcd.str(string($A,12,$B,18))
    lcd.hex(i++, 8)

PUB Font_Demo|i, j, k, c
  'set default colors
  backgroundcolor:=lcd.RGB(0,0,50) 'dark blue
  foregroundcolor:=lcd.RGB(255,255,0) 'yellow
  lcd.SetFontColors(foregroundcolor,backgroundcolor)
   
 'Change font
  lcd.SetFont(@Font1)   'Using an 8x16 font called "Progo"
 'Fill screen fast way
  lcd.FillScreen(lcd.RGB(1,64,128))

 'Draw characters all over the screen (80x30 characters) with random colors
 
  k:=0
  repeat 10
    repeat j from 0 to lcd#ypixels/16-1
      repeat i from 0 to lcd#xpixels/8-1 
        lcd.SetFontColors(lcd.RGB(?c,?c,?c),lcd.RGB(?c,?c,?c)) 
        lcd.DrawChar(i*8,j*16,?c)
        k++
        k&=$FF

PUB Scroll_Demo|i,j 'Shows how text scrolls on newline

  lcd.StartText(@font2,@lineBuffer)

  repeat 30
    repeat ((||?j)//40)
      i:= ((||?j)//100)+$E
      lcd.out(i)
    lcd.out(13)
    waitcnt(cnt+clkfreq/10)

  lcd.out(0) 'clear screen  
  lcd.ResetScroll 'reset scrolled screen to no scroll


PUB Gfx_Demo|k,i

  'Fill screen fast way
  lcd.FillScreen(lcd.RGB(1,64,128))
  'lcd.DrawRectangle(lcd#rgb_white,0,0,639,479)     'draws white boarder around whole screen
   
  'Set random pixels
  repeat 10000
    lcd.SetPixel(lcd.RGB(?k,?k,?k),||?k//lcd#xpixels,||?k//lcd#ypixels)
    'Draw random lines
  repeat 500  
    lcd.DrawLine(lcd.RGB(?k,?k,?k),||?k//lcd#xpixels,||?k//lcd#ypixels,||?k//lcd#xpixels,||?k//lcd#ypixels)
  'Draw random circles
  repeat 100
    lcd.DrawSolidCircle(lcd.RGB(?k,?k,?k),40+||?k//270,40+||?k//190,||?k//40)
  'Draw random triangles
  repeat 50
    lcd.DrawSolidTriangle(lcd.RGB(?k,?k,?k),||?k//lcd#xpixels,||?k//lcd#ypixels,||?k//lcd#xpixels,||?k//lcd#ypixels,||?k//lcd#xpixels,||?k//lcd#ypixels)
   
  'Draw random rectangles
  repeat 500
    lcd.DrawSolidRectangle(lcd.RGB(?k,?k,?k),||?k//lcd#xpixels,||?k//lcd#ypixels,||?k//lcd#xpixels,||?k//lcd#ypixels)      

  lcd.StartText(@font2,@lineBuffer)   'Driver includes an 8x12 font from VGA_HIRES_TEXT (but only characters 0..127)
  lcd.out(0) 'clear screen
  lcd.str(string("Attempting to mount SD card to show BMP images...",13))
   
  'Mounting SD card now
  repeat   'try a few times to mount the SD card
    i:=\sd.Mount_Explicit(SD_MISO,SD_CLK,SD_MOSI,SD_CS) 
    if i==0
       quit
    waitcnt(cnt+clkfreq/4)
    lcd.out(".")
 
  'Showing some 24-bit bitmap images from SD card
  k:=string("turtle.bmp")
  BMP(0,0,k)   'show the bitmap
  waitcnt(cnt+clkfreq*2)
  k:=string("garden.bmp")
  BMP(0,0,k)   'show the bitmap
  waitcnt(cnt+clkfreq*2) 



PUB BMP(bmpX,bmpY,sFilename)|i,j,k,l
'Draw a bitmap file onto the screen
    'Update speed seems to be only limited by SD card reading speed
    'note that bitmaps are stored from bottom line to top line!
    i:=\sd.popen(sFilename,"r") 'open image for reading
    if i<>0
      return i   
    sd.pread(@bfType,2) 'read bmp header
    sd.pread(@bfSize,4) 'read bmp header  
    sd.pread(@bfReserved1,4) 'read bmp header  
    sd.pread(@bfOffBits,16) 'read bmp header
    sd.pread(@biPlanes,4)
    sd.pread(@biCompression,24)

    if (biHeight<1) or (biWidth<1)
      return -1  'some kind of problem

    l:=biHeight+bmpY-1  
      
    case biBitCount
      24:  'true color (3 bytes per pixel)        
        i:=(biWidth*3)//4
        k:=biWidth*3+i     
        repeat j from l to bmpY 
          'process image row by row          
          sd.pRead(@lineBuffer,k)
          lcd.SetWindow(bmpX,constant(lcd#xpixels-1),j,constant(lcd#ypixels-1)) 
          lcd.WriteBitmapLine(@lineBuffer,biWidth,24,0)
 
      8:  '256-Palette mode
        'first need to read in palette  (rgbquad (4 bytes) * # pallete entries)
        'calculate actual bytes in palette
        i:=bfOffBits-54
        'read in palette
        sd.pread(@BmpPalette,i)
        i:=biWidth//4
        k:=biWidth+i
        repeat j from l to bmpY          
          sd.pRead(@lineBuffer,k)
          lcd.SetWindow(bmpX,constant(lcd#xpixels-1),j,constant(lcd#ypixels-1)) 
          lcd.WriteBitmapLine(@lineBuffer,biWidth,8,@BmpPalette) 
      
      4:  '16-Palette mode
        'first need to read in palette  (rgbquad (4 bytes) * # pallete entries)
        'calculate actual bytes in palette
        i:=bfOffBits-54        
        'read in palette
        sd.pread(@BmpPalette,i)
        'now paint picture
        i:=(biWidth>>1)//4
        k:=(biWidth>>1)+i 
        repeat j from l to bmpY          
          sd.pRead(@lineBuffer,k)
          lcd.SetWindow(bmpX,constant(lcd#xpixels-1),j,constant(lcd#ypixels-1)) 
          lcd.WriteBitmapLine(@lineBuffer,biWidth,4,@BmpPalette) 

    sd.pclose
    return 0



DAT     'Data for reading bitmap files

BMPHeader  byte 'Mostly using info from here:  http://www.fortunecity.com/skyscraper/windows/364/bmpffrmt.html
bfType byte "B","M"  ' 19778
bfSize long 0
bfReserved1 word 0
bfReserved2 word 0
bfOffBits long 54
biSize long 40
biWidth long 0
biHeight long 10
biPlanes word 1
biBitCount word 24
biCompression long 0
biSizeImage long 0
biXPelsPerMeter long 0
biYPelsPerMeter long 0
biClrUsed long 0
biClrImportant long 0    
DAT  'Font:  Custom font located here.  Other fonts can be selected with the SetFont command
'Note:  If you do not need a font, you can save space by commenting out the file command line below
        
font1 long
file "ProgoCo_wHeader.dat"   'nice 8x16 fixed pitch font

DAT 'Second font   
font2 byte
'file "VGA_HiRes_wHeader.dat "   'nice 8x12 copied from Parallax VGA_HiRes_Text 
'Ray's Font Editor generated font text
'Source: VGA_HiResText.rfd
'Font Name: 

'12-byte info header:
byte  "R","F","D","1"
byte  0,127,12,8  'start char, end char, height, max. width
byte  1,0,1,0  '#bytes aligned to, 1 if variable size data else 0, bpp, {bit0: if 32-bit table, bit1: if width table, bit2: if width at start of each char} 

 'Character #0
byte    $00, $00, $10, $30, $7e, $7e, $30, $10, $00, $00, $00, $00
 'Character #1
byte    $00, $00, $08, $0c, $7e, $7e, $0c, $08, $00, $00, $00, $00
 'Character #2
byte    $00, $18, $3c, $7e, $18, $18, $18, $18, $18, $18, $00, $00
 'Character #3
byte    $00, $18, $18, $18, $18, $18, $18, $7e, $3c, $18, $00, $00
 'Character #4
byte    $00, $3c, $42, $81, $81, $81, $81, $81, $42, $3c, $00, $00
 'Character #5
byte    $00, $3c, $42, $99, $bd, $bd, $bd, $99, $42, $3c, $00, $00
 'Character #6
byte    $00, $ff, $81, $81, $81, $81, $81, $81, $81, $ff, $00, $00
 'Character #7
byte    $00, $ff, $c3, $e7, $bd, $99, $bd, $e7, $c3, $ff, $00, $00
 'Character #8
byte    $40, $60, $70, $78, $7c, $7e, $7c, $78, $70, $60, $40, $00
 'Character #9
byte    $00, $00, $00, $38, $7c, $7c, $7c, $38, $00, $00, $00, $00
 'Character #10
byte    $00, $00, $00, $00, $00, $03, $0f, $0c, $18, $18, $18, $18
 'Character #11
byte    $00, $00, $00, $00, $00, $c0, $f0, $30, $18, $18, $18, $18
 'Character #12
byte    $18, $18, $18, $18, $0c, $0f, $03, $00, $00, $00, $00, $00
 'Character #13
byte    $18, $18, $18, $18, $30, $f0, $c0, $00, $00, $00, $00, $00
 'Character #14
byte    $00, $00, $00, $00, $00, $ff, $ff, $00, $00, $00, $00, $00
 'Character #15
byte    $18, $18, $18, $18, $18, $18, $18, $18, $18, $18, $18, $18
 'Character #16
byte    $00, $00, $00, $00, $00, $ff, $ff, $18, $18, $18, $18, $18
 'Character #17
byte    $18, $18, $18, $18, $18, $ff, $ff, $00, $00, $00, $00, $00
 'Character #18
byte    $18, $18, $18, $18, $18, $1f, $1f, $18, $18, $18, $18, $18
 'Character #19
byte    $18, $18, $18, $18, $18, $f8, $f8, $18, $18, $18, $18, $18
 'Character #20
byte    $18, $18, $18, $18, $18, $ff, $ff, $18, $18, $18, $18, $18
 'Character #21
byte    $00, $ff, $ff, $00, $00, $ff, $ff, $00, $00, $ff, $ff, $00
 'Character #22
byte    $66, $cc, $99, $33, $66, $cc, $99, $33, $66, $cc, $99, $33
 'Character #23
byte    $66, $66, $66, $66, $66, $66, $66, $66, $66, $66, $66, $66
 'Character #24
byte    $aa, $55, $aa, $55, $aa, $55, $aa, $55, $aa, $55, $aa, $55
 'Character #25
byte    $f0, $f0, $f0, $f0, $f0, $f0, $ff, $ff, $ff, $ff, $ff, $ff
 'Character #26
byte    $f0, $f0, $f0, $f0, $f0, $f0, $0f, $0f, $0f, $0f, $0f, $0f
 'Character #27
byte    $f0, $f0, $f0, $f0, $f0, $f0, $f0, $f0, $f0, $f0, $f0, $f0
 'Character #28
byte    $f0, $f0, $f0, $f0, $f0, $f0, $00, $00, $00, $00, $00, $00
 'Character #29
byte    $00, $00, $00, $00, $00, $00, $ff, $ff, $ff, $ff, $ff, $ff
 'Character #30
byte    $00, $00, $00, $00, $00, $00, $0f, $0f, $0f, $0f, $0f, $0f
 'Character #31
byte    $00, $00, $00, $00, $00, $00, $f0, $f0, $f0, $f0, $f0, $f0
 'Character #32
byte    $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00
 'Character #33
byte    $00, $18, $3c, $3c, $3c, $18, $18, $00, $18, $18, $00, $00
 'Character #34
byte    $00, $66, $66, $ee, $cc, $00, $00, $00, $00, $00, $00, $00
 'Character #35
byte    $00, $6c, $6c, $fe, $6c, $6c, $6c, $fe, $6c, $6c, $00, $00
 'Character #36
byte    $18, $18, $3e, $66, $60, $3c, $06, $66, $7c, $18, $18, $00
 'Character #37
byte    $00, $00, $00, $62, $66, $0c, $18, $30, $66, $46, $00, $00
 'Character #38
byte    $00, $70, $d8, $d8, $70, $da, $de, $cc, $dc, $76, $00, $00
 'Character #39
byte    $00, $18, $18, $38, $30, $00, $00, $00, $00, $00, $00, $00
 'Character #40
byte    $00, $0c, $18, $30, $60, $60, $60, $30, $18, $0c, $00, $00
 'Character #41
byte    $00, $60, $30, $18, $0c, $0c, $0c, $18, $30, $60, $00, $00
 'Character #42
byte    $00, $00, $00, $66, $3c, $ff, $3c, $66, $00, $00, $00, $00
 'Character #43
byte    $00, $00, $00, $18, $18, $7e, $18, $18, $00, $00, $00, $00
 'Character #44
byte    $00, $00, $00, $00, $00, $00, $00, $00, $38, $38, $18, $30
 'Character #45
byte    $00, $00, $00, $00, $00, $7e, $00, $00, $00, $00, $00, $00
 'Character #46
byte    $00, $00, $00, $00, $00, $00, $00, $00, $38, $38, $00, $00
 'Character #47
byte    $00, $00, $02, $06, $0c, $18, $30, $60, $c0, $80, $00, $00
 'Character #48
byte    $00, $7c, $c6, $ce, $de, $d6, $f6, $e6, $c6, $7c, $00, $00
 'Character #49
byte    $00, $08, $18, $78, $18, $18, $18, $18, $18, $7e, $00, $00
 'Character #50
byte    $00, $3c, $66, $66, $06, $0c, $18, $30, $66, $7e, $00, $00
 'Character #51
byte    $00, $3c, $66, $06, $06, $1c, $06, $06, $66, $3c, $00, $00
 'Character #52
byte    $00, $0c, $1c, $3c, $6c, $cc, $fe, $0c, $0c, $1e, $00, $00
 'Character #53
byte    $00, $7e, $60, $60, $60, $7c, $06, $06, $66, $3c, $00, $00
 'Character #54
byte    $00, $1c, $30, $60, $60, $7c, $66, $66, $66, $3c, $00, $00
 'Character #55
byte    $00, $fe, $c6, $c6, $06, $0c, $18, $30, $30, $30, $00, $00
 'Character #56
byte    $00, $3c, $66, $66, $76, $3c, $6e, $66, $66, $3c, $00, $00
 'Character #57
byte    $00, $3c, $66, $66, $66, $3e, $06, $06, $0c, $38, $00, $00
 'Character #58
byte    $00, $00, $00, $38, $38, $00, $00, $38, $38, $00, $00, $00
 'Character #59
byte    $00, $00, $00, $00, $38, $38, $00, $00, $38, $38, $18, $30
 'Character #60
byte    $00, $06, $0c, $18, $30, $60, $30, $18, $0c, $06, $00, $00
 'Character #61
byte    $00, $00, $00, $00, $7e, $00, $7e, $00, $00, $00, $00, $00
 'Character #62
byte    $00, $60, $30, $18, $0c, $06, $0c, $18, $30, $60, $00, $00
 'Character #63
byte    $00, $3c, $66, $06, $0c, $18, $18, $00, $18, $18, $00, $00
 'Character #64
byte    $00, $7c, $e6, $c6, $de, $de, $dc, $c0, $e0, $7c, $00, $00
 'Character #65
byte    $00, $3c, $66, $66, $66, $7e, $66, $66, $66, $66, $00, $00
 'Character #66
byte    $00, $fc, $66, $66, $66, $7c, $66, $66, $66, $fc, $00, $00
 'Character #67
byte    $00, $3c, $66, $c6, $c0, $c0, $c0, $c6, $66, $3c, $00, $00
 'Character #68
byte    $00, $f8, $6c, $66, $66, $66, $66, $66, $6c, $f8, $00, $00
 'Character #69
byte    $00, $fe, $62, $60, $64, $7c, $64, $60, $62, $fe, $00, $00
 'Character #70
byte    $00, $fe, $62, $60, $64, $7c, $64, $60, $60, $f0, $00, $00
 'Character #71
byte    $00, $3c, $66, $c6, $c0, $c0, $ce, $c6, $66, $3e, $00, $00
 'Character #72
byte    $00, $c6, $c6, $c6, $c6, $fe, $c6, $c6, $c6, $c6, $00, $00
 'Character #73
byte    $00, $3c, $18, $18, $18, $18, $18, $18, $18, $3c, $00, $00
 'Character #74
byte    $00, $1e, $0c, $0c, $0c, $0c, $cc, $cc, $cc, $78, $00, $00
 'Character #75
byte    $00, $e6, $66, $6c, $6c, $78, $6c, $6c, $66, $e6, $00, $00
 'Character #76
byte    $00, $f0, $60, $60, $60, $60, $62, $66, $66, $fe, $00, $00
 'Character #77
byte    $00, $c6, $ee, $fe, $fe, $d6, $c6, $c6, $c6, $c6, $00, $00
 'Character #78
byte    $00, $c6, $c6, $e6, $f6, $fe, $de, $ce, $c6, $c6, $00, $00
 'Character #79
byte    $00, $38, $6c, $c6, $c6, $c6, $c6, $c6, $6c, $38, $00, $00
 'Character #80
byte    $00, $fc, $66, $66, $66, $7c, $60, $60, $60, $f0, $00, $00
 'Character #81
byte    $00, $38, $6c, $c6, $c6, $c6, $c6, $de, $6c, $3c, $06, $00
 'Character #82
byte    $00, $fc, $66, $66, $66, $7c, $6c, $66, $66, $e6, $00, $00
 'Character #83
byte    $00, $3c, $66, $66, $60, $38, $0c, $66, $66, $3c, $00, $00
 'Character #84
byte    $00, $7e, $5a, $18, $18, $18, $18, $18, $18, $3c, $00, $00
 'Character #85
byte    $00, $66, $66, $66, $66, $66, $66, $66, $66, $3c, $00, $00
 'Character #86
byte    $00, $66, $66, $66, $66, $66, $66, $66, $3c, $18, $00, $00
 'Character #87
byte    $00, $c6, $c6, $c6, $c6, $d6, $d6, $6c, $6c, $6c, $00, $00
 'Character #88
byte    $00, $66, $66, $66, $3c, $18, $3c, $66, $66, $66, $00, $00
 'Character #89
byte    $00, $66, $66, $66, $66, $3c, $18, $18, $18, $3c, $00, $00
 'Character #90
byte    $00, $fe, $c6, $8c, $18, $30, $60, $c2, $c6, $fe, $00, $00
 'Character #91
byte    $00, $3c, $30, $30, $30, $30, $30, $30, $30, $3c, $00, $00
 'Character #92
byte    $00, $00, $80, $c0, $60, $30, $18, $0c, $06, $02, $00, $00
 'Character #93
byte    $00, $3c, $0c, $0c, $0c, $0c, $0c, $0c, $0c, $3c, $00, $00
 'Character #94
byte    $00, $10, $38, $6c, $c6, $00, $00, $00, $00, $00, $00, $00
 'Character #95
byte    $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $ff, $ff
 'Character #96
byte    $00, $00, $00, $30, $38, $1c, $0c, $00, $00, $00, $00, $00
 'Character #97
byte    $00, $00, $00, $00, $78, $0c, $7c, $cc, $cc, $76, $00, $00
 'Character #98
byte    $00, $e0, $60, $60, $7c, $66, $66, $66, $66, $dc, $00, $00
 'Character #99
byte    $00, $00, $00, $00, $3c, $66, $60, $60, $66, $3c, $00, $00
 'Character #100
byte    $00, $1c, $0c, $0c, $7c, $cc, $cc, $cc, $cc, $76, $00, $00
 'Character #101
byte    $00, $00, $00, $00, $3c, $66, $7e, $60, $66, $3c, $00, $00
 'Character #102
byte    $00, $1c, $36, $30, $30, $7c, $30, $30, $30, $78, $00, $00
 'Character #103
byte    $00, $00, $00, $00, $76, $cc, $cc, $cc, $7c, $0c, $cc, $78
 'Character #104
byte    $00, $e0, $60, $60, $6c, $76, $66, $66, $66, $e6, $00, $00
 'Character #105
byte    $00, $18, $18, $00, $38, $18, $18, $18, $18, $7e, $00, $00
 'Character #106
byte    $00, $06, $06, $00, $0e, $06, $06, $06, $06, $66, $66, $3c
 'Character #107
byte    $00, $e0, $60, $60, $66, $6c, $78, $6c, $66, $e6, $00, $00
 'Character #108
byte    $00, $78, $18, $18, $18, $18, $18, $18, $18, $7e, $00, $00
 'Character #109
byte    $00, $00, $00, $00, $fc, $d6, $d6, $d6, $d6, $d6, $00, $00
 'Character #110
byte    $00, $00, $00, $00, $7c, $66, $66, $66, $66, $66, $00, $00
 'Character #111
byte    $00, $00, $00, $00, $3c, $66, $66, $66, $66, $3c, $00, $00
 'Character #112
byte    $00, $00, $00, $00, $dc, $66, $66, $66, $66, $7c, $60, $f0
 'Character #113
byte    $00, $00, $00, $00, $76, $cc, $cc, $cc, $cc, $7c, $0c, $1e
 'Character #114
byte    $00, $00, $00, $00, $ec, $6e, $76, $60, $60, $f0, $00, $00
 'Character #115
byte    $00, $00, $00, $00, $3c, $66, $30, $0c, $66, $3c, $00, $00
 'Character #116
byte    $00, $00, $10, $30, $7e, $30, $30, $30, $36, $1c, $00, $00
 'Character #117
byte    $00, $00, $00, $00, $cc, $cc, $cc, $cc, $cc, $76, $00, $00
 'Character #118
byte    $00, $00, $00, $00, $66, $66, $66, $66, $3c, $18, $00, $00
 'Character #119
byte    $00, $00, $00, $00, $c6, $c6, $d6, $d6, $6c, $6c, $00, $00
 'Character #120
byte    $00, $00, $00, $00, $c6, $6c, $38, $38, $6c, $c6, $00, $00
 'Character #121
byte    $00, $00, $00, $00, $66, $66, $66, $66, $3e, $06, $0c, $38
 'Character #122
byte    $00, $00, $00, $00, $7e, $46, $0c, $30, $62, $7e, $00, $00
 'Character #123
byte    $00, $0e, $18, $18, $30, $60, $30, $18, $18, $0e, $00, $00
 'Character #124
byte    $00, $18, $18, $18, $18, $18, $18, $18, $18, $18, $00, $00
 'Character #125
byte    $00, $70, $18, $18, $0c, $06, $0c, $18, $18, $70, $00, $00
 'Character #126
byte    $00, $73, $db, $ce, $00, $00, $00, $00, $00, $00, $00, $00
 'Character #127
byte    $00, $00, $18, $18, $7e, $18, $18, $00, $7e, $00, $00, $00

DAT 'MIT License
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}